use std::cell::{Cell, OnceCell, RefCell};

use adw::{prelude::*, subclass::prelude::*};
use atspi::{proxy::accessible::ObjectRefExt, ObjectRef};
use glib::Properties;
use gtk::{gio, glib};

use crate::{
    bus,
    i18n::gettext,
    traits::{Iconic, Translatable},
};

mod imp {
    use super::*;

    #[derive(Debug, Default, Properties)]
    #[properties(wrapper_type = super::ElevadoNode)]
    pub struct ElevadoNode {
        pub(super) accessible: OnceCell<ObjectRef>,
        pub(super) role: Cell<Option<atspi::Role>>,
        pub(super) state: Cell<atspi::StateSet>,
        pub(super) children: RefCell<Vec<super::ElevadoNode>>,

        #[property(name = "role-name", type = String, get = Self::role_name)]
        #[property(name = "role-icon-name", type = String, get = |imp: &Self| String::from(imp.role.get().unwrap_or(atspi::Role::Invalid).iconify()))]
        // Name will be None if call failed or string is empty
        #[property(get)]
        pub(super) name: RefCell<Option<String>>,
        #[property(get)]
        pub(super) object_path: RefCell<String>,
        #[property(get)]
        pub(super) n_children: Cell<u32>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for ElevadoNode {
        type Interfaces = (gio::ListModel,);
        type Type = super::ElevadoNode;

        const NAME: &'static str = "ElevadoNode";
    }

    #[glib::derived_properties]
    impl ObjectImpl for ElevadoNode {}
    impl ListModelImpl for ElevadoNode {
        fn n_items(&self) -> u32 {
            self.n_children.get()
        }

        fn item_type(&self) -> glib::Type {
            super::ElevadoNode::static_type()
        }

        fn item(&self, position: u32) -> Option<glib::Object> {
            if (position + 1) > self.n_children.get() {
                return None;
            }

            self.children
                .borrow()
                .get(position as usize)
                .map(|i| i.clone().upcast())
        }
    }

    impl ElevadoNode {
        pub(super) async fn set_accessible(&self, accessible: ObjectRef) {
            self.accessible
                .set(accessible)
                .expect("accessible should never be written to twice");
            self.pull_props_from_accessible(false).await;
        }

        #[expect(clippy::cast_sign_loss)]
        async fn pull_props_from_accessible(&self, notify: bool) {
            if notify {
                let old_n_children = self.n_children.get();
                self.n_children.set(0);
                self.children.borrow_mut().clear();

                self.obj().items_changed(0, old_n_children, 0);
            }

            let objref = self
                .accessible
                .get()
                .expect("accessible must be set already");
            let proxy = objref
                .as_accessible_proxy(bus().connection())
                .await
                .expect("all objects must implement accessible");

            let n = proxy.name().await.unwrap_or_default();
            let name = if n.is_empty() { None } else { Some(n) };
            self.name.replace(name);

            self.object_path.replace(objref.path.as_str().to_string());

            self.state.set(proxy.get_state().await.unwrap_or_default());
            self.role.set(proxy.get_role().await.ok());

            let n: i32 = proxy.child_count().await.unwrap_or_default();
            self.n_children.set(n as u32);

            if notify {
                let nobj = self.obj();
                let _ = nobj.freeze_notify();

                nobj.notify_name();
                nobj.notify_n_children();
                nobj.notify_object_path();
                nobj.notify_role_name();
                nobj.notify_role_icon_name();
            }
        }

        fn role_name(&self) -> String {
            if let Some(role) = self.role.get() {
                role.translate()
            } else {
                gettext("No Role")
            }
        }
    }
}

glib::wrapper! {
    pub struct ElevadoNode(ObjectSubclass<imp::ElevadoNode>)
        @implements gio::ListModel;
}

impl ElevadoNode {
    pub async fn new(accessible: ObjectRef) -> Self {
        let node: Self = glib::Object::new();
        node.imp().set_accessible(accessible).await;

        node
    }

    pub fn accessible(&self) -> Option<&ObjectRef> {
        self.imp().accessible.get()
    }

    pub fn role(&self) -> Option<atspi::Role> {
        self.imp().role.get()
    }

    pub fn state(&self) -> atspi::StateSet {
        self.imp().state.get()
    }

    pub async fn init_children(&self) {
        if !self.imp().children.borrow().is_empty() {
            return;
        }

        let obj = self.accessible().expect("accessible should already be set");

        let ap = obj
            .as_accessible_proxy(bus().connection())
            .await
            .expect("all objects implement accessible");

        let child_objs: Vec<ObjectRef> = ap.get_children().await.unwrap_or_default();
        let mut nodes: Vec<ElevadoNode> = vec![];
        for obj in child_objs {
            let node = ElevadoNode::new(obj).await;
            nodes.push(node);
        }
        self.imp().children.replace(nodes);
    }
}

impl Default for ElevadoNode {
    fn default() -> Self {
        glib::Object::new()
    }
}
