use std::{cell::RefCell, error::Error, time::Duration};

use adw::{prelude::*, subclass::prelude::*};
use atspi::{
    events::ObjectEvents,
    proxy::{
        accessible::{AccessibleProxy, ObjectRefExt},
        action::ActionProxy,
        application::ApplicationProxy,
        component::ComponentProxy,
        proxy_ext::ProxyExt,
        text::TextProxy,
        value::ValueProxy,
    },
    zbus::fdo::DBusProxy,
    EventProperties, ObjectRef, RelationType,
};
use glib::{clone, Properties};
use gtk::{glib, pango};
use smol::stream::StreamExt;

use crate::{
    bus,
    i18n::{gettext, ngettext_f},
    model::ElevadoNode,
    traits::Translatable,
};

mod imp {
    use super::*;

    #[derive(Debug, Default, Properties, gtk::CompositeTemplate)]
    #[properties(wrapper_type = super::ElevadoNodeDetails)]
    #[template(resource = "/com/feaneron/Elevado/panels/details/node_details.ui")]
    pub struct ElevadoNodeDetails {
        pub(super) relation_rows: RefCell<Vec<adw::ActionRow>>,
        pub(super) action_rows: RefCell<Vec<adw::ActionRow>>,

        #[property(get, set = Self::set_node)]
        pub(super) node: RefCell<ElevadoNode>,

        #[template_child]
        pub(super) view_stack: TemplateChild<adw::ViewStack>,
        #[template_child]
        pub(super) error_output: TemplateChild<gtk::Label>,

        #[template_child]
        pub(super) name_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) description_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) state_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) interfaces_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) help_text_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) index_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) role_row: TemplateChild<adw::ActionRow>,

        #[template_child]
        pub(super) component_group: TemplateChild<adw::PreferencesGroup>,
        #[template_child]
        pub(super) component_size_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) component_position_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) component_layer_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) component_alpha_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) component_mdi_z_order_row: TemplateChild<adw::ActionRow>,

        #[template_child]
        pub(super) toolkit_group: TemplateChild<adw::PreferencesGroup>,
        #[template_child]
        pub(super) toolkit_name_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) toolkit_version_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) toolkit_atspi_version_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) toolkit_application_id_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) toolkit_process_id_row: TemplateChild<adw::ActionRow>,

        #[template_child]
        pub(super) value_group: TemplateChild<adw::PreferencesGroup>,
        #[template_child]
        pub(super) current_value_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) minimum_value_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) maximum_value_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) minimum_increment_value_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) text_value_row: TemplateChild<adw::ActionRow>,

        #[template_child]
        pub(super) text_group: TemplateChild<adw::PreferencesGroup>,
        #[template_child]
        pub(super) text_row: TemplateChild<adw::ActionRow>,
        #[template_child]
        pub(super) caret_row: TemplateChild<adw::ActionRow>,

        #[template_child]
        pub(super) action_group: TemplateChild<adw::PreferencesGroup>,
        #[template_child]
        pub(super) action_spinner: TemplateChild<adw::Spinner>,
        #[template_child]
        pub(super) relations_group: TemplateChild<adw::PreferencesGroup>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for ElevadoNodeDetails {
        type ParentType = adw::Bin;
        type Type = super::ElevadoNodeDetails;

        const NAME: &'static str = "ElevadoNodeDetails";

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    #[glib::derived_properties]
    impl ObjectImpl for ElevadoNodeDetails {
        fn constructed(&self) {
            glib::spawn_future_local(clone!(
                #[weak(rename_to = this)]
                self,
                async move {
                    this.monitor_events().await;
                }
            ));
        }
    }
    impl WidgetImpl for ElevadoNodeDetails {}
    impl BinImpl for ElevadoNodeDetails {}

    impl ElevadoNodeDetails {
        fn set_node(&self, value: ElevadoNode) {
            self.node.replace(value);
            glib::spawn_future_local(clone!(
                #[weak(rename_to = this)]
                self,
                async move {
                    this.update_details().await;
                }
            ));
        }

        async fn monitor_events(&self) {
            // This will be deregistered in main before exit.
            bus()
                .register_event::<ObjectEvents>()
                .await
                .expect("registering events shouldn't fail");
            let mut stream = bus().event_stream();
            std::pin::pin!(&mut stream);

            // Even if multiple events come at the same time, update the GUI only once
            let (tx, rx) = smol::channel::bounded(1);

            // zbus stream must always be polled or else it will hang
            // so we do so in a separate thread
            smol::unblock(|| {
                smol::block_on(async move {
                    while let Some(Ok(event)) = stream.next().await {
                        // If receiver is full, we don't want to wait
                        let _ = tx.try_send(event);
                    }
                });
            })
            .detach();

            // If the event is about this node, refresh GUI
            while let Ok(event) = rx.recv().await {
                if let Some(obj) = self.obj().node().accessible() {
                    if event.object_ref() == *obj {
                        self.update_details().await;
                    }
                }
            }
        }

        async fn update_details(&self) {
            self.view_stack.set_visible_child_name("main-page");
            if let Err(e) = self.update_groups().await {
                self.view_stack.set_visible_child_name("error-page");
                self.error_output.set_label(&e.to_string());
            }
        }

        async fn update_groups(&self) -> Result<(), Box<dyn Error>> {
            let node = self.obj().node();
            let objref = node
                .accessible()
                .ok_or("Elevado: node accessible is not set")?;

            let ap = objref.as_accessible_proxy(bus().connection()).await?;

            {
                // Accessible Group
                let name = ap.name().await.unwrap_or_default();
                self.name_row.set_visible(!name.is_empty());
                self.name_row.set_subtitle(&name);

                let description = ap.description().await.unwrap_or_default();
                self.description_row.set_visible(!description.is_empty());
                self.description_row.set_subtitle(&description);

                let help_text = ap.help_text().await.unwrap_or_default();
                self.help_text_row.set_visible(!help_text.is_empty());
                self.help_text_row.set_subtitle(&help_text);

                let interfaces = ap.get_interfaces().await.unwrap_or_default();
                let iface_names: String = interfaces
                    .iter()
                    .map(|i| i.translate())
                    .collect::<Vec<String>>()
                    .join(", ");
                self.interfaces_row.set_visible(!iface_names.is_empty());
                self.interfaces_row.set_subtitle(&iface_names);

                let states = ap.get_state().await.unwrap_or_default();
                self.state_row.set_visible(!states.is_empty());
                {
                    let names: String = states
                        .iter()
                        .map(|s| s.translate())
                        .collect::<Vec<String>>()
                        .join(", ");
                    self.state_row.set_subtitle(&names);
                }

                self.role_row.set_subtitle(&node.role_name());

                self.index_row.set_subtitle(
                    &ap.get_index_in_parent()
                        .await
                        .unwrap_or_default()
                        .to_string(),
                );
            }

            // Convert between proxies easily. Returns err if interface isn't available
            let proxyswitch = ap.proxies().await?;

            // Toolkit (Application) Group
            if let Ok(app_proxy) = proxyswitch.application().await {
                self.update_toolkit_group(app_proxy).await;
            } else {
                self.toolkit_group.set_visible(false);
            }

            // Component Group
            if let Ok(comp_proxy) = proxyswitch.component().await {
                self.update_component_group(comp_proxy).await;
            } else {
                self.component_group.set_visible(false);
            }

            // Text Group
            if let Ok(text_proxy) = proxyswitch.text().await {
                self.update_text_group(text_proxy).await;
            } else {
                self.text_group.set_visible(false);
            }

            // Value Group
            if let Ok(value_proxy) = proxyswitch.value().await {
                self.update_value_group(value_proxy).await;
            } else {
                self.value_group.set_visible(false);
            }

            // Action Group
            if let Ok(action_proxy) = proxyswitch.action().await {
                self.update_action_group(action_proxy).await;
            } else {
                self.action_group.set_visible(false);
            }

            // Relation Group
            self.update_relations_group(&ap).await;

            Ok(())
        }

        async fn update_relations_group(&self, accessible: &AccessibleProxy<'_>) {
            {
                // Drop mut ref before await
                let mut relation_row_list = self.relation_rows.borrow_mut();
                for row in relation_row_list.iter() {
                    self.relations_group.remove(row);
                }
                relation_row_list.clear();
            }

            let relations: Vec<(RelationType, Vec<ObjectRef>)> =
                accessible.get_relation_set().await.unwrap_or_default();
            if relations.is_empty() {
                self.relations_group.set_visible(false);
                return;
            }

            for relation in relations {
                let row = adw::ActionRow::new();
                row.set_title(&relation.0.translate());
                let n_targets = u32::try_from(relation.1.len())
                    .expect("relation length should convert into u32");
                row.set_subtitle(
                    // Translators: Do not translate the content between the braces {}
                    &ngettext_f(
                        "{n} Object",
                        "{n} Objects",
                        n_targets,
                        &[("n", &n_targets.to_string())],
                    ),
                );

                for (i, target) in relation.1.iter().enumerate() {
                    let mut name = target
                        .as_accessible_proxy(bus().connection())
                        .await
                        .expect("all objects must implement accessible")
                        .name()
                        .await
                        .unwrap_or_default();
                    if name.is_empty() {
                        name = gettext("Unnamed");
                    }

                    let label = gtk::Label::builder()
                        .label(&name)
                        .ellipsize(pango::EllipsizeMode::End)
                        .build();

                    let button = gtk::Button::builder()
                        .child(&label)
                        .tooltip_text(&name)
                        .valign(gtk::Align::Center)
                        .has_frame(false)
                        .build();
                    row.add_suffix(&button);
                    if i > 0 {
                        button
                            .parent()
                            .and_dynamic_cast::<gtk::Orientable>()
                            .expect("parent widget should implement Orientable iface")
                            .set_orientation(gtk::Orientation::Vertical);
                    }
                }

                self.relations_group.add(&row);
                self.relation_rows.borrow_mut().push(row);
            }
            self.relation_rows.borrow_mut().shrink_to_fit();

            self.relations_group.set_visible(true);
        }

        async fn update_toolkit_group(&self, app_proxy: ApplicationProxy<'_>) {
            self.toolkit_name_row
                .set_subtitle(&app_proxy.toolkit_name().await.unwrap_or_default());
            self.toolkit_version_row
                .set_subtitle(&app_proxy.version().await.unwrap_or_default());
            self.toolkit_atspi_version_row
                .set_subtitle(&app_proxy.atspi_version().await.unwrap_or_default());
            self.toolkit_application_id_row
                .set_subtitle(&app_proxy.id().await.unwrap_or_default().to_string());

            let busname = app_proxy.inner().destination();
            let dproxy = DBusProxy::from(app_proxy.inner().clone());
            if let Ok(proc_id) = dproxy.get_connection_unix_process_id(busname.into()).await {
                self.toolkit_process_id_row
                    .set_subtitle(&proc_id.to_string());
                self.toolkit_process_id_row.set_visible(true);
            } else {
                self.toolkit_process_id_row.set_visible(false);
            }

            self.toolkit_group.set_visible(true);
        }

        async fn update_action_group(&self, action_proxy: ActionProxy<'_>) {
            let actions = action_proxy.get_actions().await.unwrap_or_default();
            if actions.is_empty() {
                self.action_group.set_visible(false);
                return;
            }

            {
                // Empty out the prior Action widgets
                let mut rows = self.action_rows.borrow_mut();
                for old_row in rows.iter() {
                    self.action_group.remove(old_row);
                }
                rows.clear();
            }

            self.action_group.set_visible(true);
            self.action_spinner.set_visible(true);

            // Create and add the new ones
            for action in actions {
                let row = adw::ActionRow::builder()
                    .title(&action.name)
                    .subtitle(&action.description)
                    .build();

                for key in action.keybinding.split(";;") {
                    row.add_suffix(&gtk::Label::new(Some(key)));
                }

                // Stagger work to prevent GUI from locking up
                glib::timeout_future(Duration::from_millis(20)).await;

                self.action_group.add(&row);
                self.action_rows.borrow_mut().push(row);
            }
            self.action_spinner.set_visible(false);
        }

        async fn update_component_group(&self, component_proxy: ComponentProxy<'_>) {
            let position = component_proxy
                .get_position(atspi::CoordType::Window)
                .await
                .expect("shouldn't fail if object implements Component");
            let size = component_proxy
                .get_size()
                .await
                .expect("shouldn't fail if object implements Component");

            self.component_position_row
                .set_subtitle(&format!("({0}, {1})", position.0, position.1));
            self.component_size_row
                .set_subtitle(&format!("({0}, {1})", size.0, size.1));
            self.component_mdi_z_order_row.set_subtitle(
                &component_proxy
                    .get_mdiz_order()
                    .await
                    .expect("shouldn't fail if object implements Component")
                    .to_string(),
            );
            self.component_alpha_row.set_subtitle(
                &component_proxy
                    .get_alpha()
                    .await
                    .expect("shouldn't fail if object implements Component")
                    .to_string(),
            );

            let layer_name = component_proxy
                .get_layer()
                .await
                .expect("shouldn't fail if object implements Component")
                .translate();
            self.component_layer_row.set_subtitle(&layer_name);

            self.component_group.set_visible(true);
        }

        async fn update_text_group(&self, text_proxy: TextProxy<'_>) {
            self.caret_row.set_subtitle(
                &text_proxy
                    .caret_offset()
                    .await
                    .unwrap_or_default()
                    .to_string(),
            );

            let char_count = text_proxy
                .character_count()
                .await
                .expect("shouldn't fail if obj has Text");
            self.text_row
                .set_subtitle(&text_proxy.get_text(0, char_count).await.unwrap_or_default());

            self.text_group.set_visible(true);
        }

        async fn update_value_group(&self, value_proxy: ValueProxy<'_>) {
            self.text_value_row.set_subtitle(
                &value_proxy
                    .text()
                    .await
                    .expect("shouldn't fail if obj has Value"),
            );

            self.current_value_row.set_subtitle(
                &value_proxy
                    .current_value()
                    .await
                    .expect("shouldn't fail if obj has Value")
                    .to_string(),
            );
            self.minimum_value_row.set_subtitle(
                &value_proxy
                    .minimum_value()
                    .await
                    .expect("shouldn't fail if obj has Value")
                    .to_string(),
            );
            self.maximum_value_row.set_subtitle(
                &value_proxy
                    .maximum_value()
                    .await
                    .expect("shouldn't fail if obj has Value")
                    .to_string(),
            );
            self.minimum_increment_value_row.set_subtitle(
                &value_proxy
                    .minimum_increment()
                    .await
                    .expect("shouldn't fail if obj has Value")
                    .to_string(),
            );

            self.value_group.set_visible(true);
        }
    }
}

glib::wrapper! {
    pub struct ElevadoNodeDetails(ObjectSubclass<imp::ElevadoNodeDetails>)
        @extends gtk::Widget, adw::Bin,
        @implements gtk::Accessible, gtk::Buildable, gtk::ConstraintTarget;
}
