use std::{cell::RefCell, sync::OnceLock};

use adw::{prelude::*, subclass::prelude::*};
use glib::{clone, subclass::Signal, Properties};
use gtk::{gdk, gio, glib};

use super::ElevadoNodeRow;
use crate::model::ElevadoNode;

mod imp {
    use super::*;

    #[derive(Debug, Default, Properties, gtk::CompositeTemplate)]
    #[properties(wrapper_type = super::ElevadoNodeTree)]
    #[template(resource = "/com/feaneron/Elevado/view/node_tree.ui")]
    pub struct ElevadoNodeTree {
        #[property(get, set = Self::set_root)]
        pub(super) root: RefCell<ElevadoNode>,

        #[template_child]
        pub(super) column_view: TemplateChild<gtk::ColumnView>,
        #[template_child]
        pub(super) single_selection: TemplateChild<gtk::SingleSelection>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for ElevadoNodeTree {
        type ParentType = adw::Bin;
        type Type = super::ElevadoNodeTree;

        const NAME: &'static str = "ElevadoNodeTree";

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
            ElevadoNodeRow::ensure_type();
            ElevadoNode::ensure_type();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    #[glib::derived_properties]
    impl ObjectImpl for ElevadoNodeTree {
        fn constructed(&self) {
            self.parent_constructed();

            // Shortcut: Use arrow keys to open/close row expander
            let controller = gtk::ShortcutController::new();
            controller
                .dynamic_cast_ref::<gtk::EventController>()
                .expect("controller should be castable to EventController")
                .set_propagation_phase(gtk::PropagationPhase::Capture);

            for key in [gdk::Key::Left, gdk::Key::Right] {
                let trigger = gtk::KeyvalTrigger::new(key, gdk::ModifierType::NO_MODIFIER_MASK);
                let cb =
                    gtk::CallbackAction::new(move |widget: &gtk::Widget, _| -> glib::Propagation {
                        widget
                            .downcast_ref::<super::ElevadoNodeTree>()
                            .expect("must be a NodeTree")
                            .imp()
                            .expander_shortcut_cb(key)
                    });

                let shortcut = gtk::Shortcut::builder()
                    .trigger(&trigger)
                    .action(&cb)
                    .build();
                controller.add_shortcut(shortcut);
            }
            self.obj().add_controller(controller);

            // On clicking NodeRow, set the selected node for NodeDetails panel
            self.column_view.connect_activate(clone!(
                #[weak(rename_to = tree)]
                self,
                move |colview: &gtk::ColumnView, _| {
                    let node = colview
                        .model()
                        .and_downcast::<gtk::SingleSelection>()
                        .expect("should be a SingleSelection")
                        .selected_item()
                        .and_downcast::<gtk::TreeListRow>()
                        .expect("this should be a TreeListRow")
                        .item()
                        .and_downcast::<ElevadoNode>()
                        .expect("item should be an ElevadoNode");

                    tree.obj().emit_by_name::<()>("selected", &[&node]);
                }
            ));
        }

        fn signals() -> &'static [Signal] {
            static SIGNALS: OnceLock<Vec<Signal>> = OnceLock::new();
            SIGNALS.get_or_init(|| {
                vec![Signal::builder("selected")
                    .run_last()
                    .param_types([ElevadoNode::static_type()])
                    .build()]
            })
        }
    }
    impl WidgetImpl for ElevadoNodeTree {}
    impl BinImpl for ElevadoNodeTree {}

    #[gtk::template_callbacks]
    impl ElevadoNodeTree {
        fn expander_shortcut_cb(&self, keyval: gdk::Key) -> glib::Propagation {
            use glib::Propagation;
            use gtk::{ListScrollFlags, TextDirection};

            // HACK: Get focused NodeRow
            let mut child = self
                .obj()
                .focus_child()
                .expect("NodeTree must have a child focused")
                .root()
                .expect("root window must exist")
                .focus()
                .expect("root should have a focus");

            let noderow: ElevadoNodeRow = loop {
                if let Some(t) = child.first_child() {
                    child = t;
                    if let Some(noderow) = child.dynamic_cast_ref::<ElevadoNodeRow>() {
                        break noderow.clone();
                    }
                } else {
                    return Propagation::Proceed;
                }
            };
            let row = noderow.row().expect("NodeRow should have a row");
            // END OF HACK

            let direction = self.obj().direction();
            let collapse: bool = (keyval == gdk::Key::Left && direction == TextDirection::Ltr)
                || (keyval == gdk::Key::Right && direction == TextDirection::Rtl);

            if !row.is_expanded() && collapse {
                if let Some(parent) = row.parent() {
                    let parent_position = parent.position();

                    self.column_view.scroll_to(
                        parent_position,
                        None,
                        ListScrollFlags::FOCUS | ListScrollFlags::SELECT,
                        None,
                    );
                }
            } else if collapse {
                noderow.collapse();
            } else {
                noderow.expand();
            }

            Propagation::Stop
        }

        fn set_root(&self, value: Option<ElevadoNode>) {
            // Refresh TreeModel on changes to the root Node
            if let Some(newroot) = value {
                if newroot != *self.root.borrow() {
                    self.root.replace(newroot);
                }
                let root = self.root.borrow();
                let store = gio::ListStore::new::<ElevadoNode>();
                store.append(&root.clone());

                let treemodel =
                    gtk::TreeListModel::new(store, false, false, |obj: &glib::Object| {
                        let node = obj.downcast_ref::<ElevadoNode>().expect("should be node");
                        if node.n_items() == 0 {
                            None
                        } else {
                            glib::spawn_future_local(clone!(
                                #[weak]
                                node,
                                async move {
                                    node.init_children().await;
                                }
                            ));
                            Some(node.clone().upcast::<gio::ListModel>())
                        }
                    });
                self.single_selection.set_model(Some(&treemodel));
            }
        }
    }
}

glib::wrapper! {
    pub struct ElevadoNodeTree(ObjectSubclass<imp::ElevadoNodeTree>)
        @extends gtk::Widget, adw::Bin,
        @implements gtk::Accessible, gtk::Buildable, gtk::ConstraintTarget;
}
