use adw::subclass::prelude::*;
use atspi::proxy::accessible::AccessibleProxy;
use glib::clone;
use gtk::{gio, glib, prelude::*};

use crate::{
    bus,
    i18n::gettext_f,
    model::ElevadoNode,
    panels::{ElevadoNodeDetails, ElevadoSidebarRow},
    view::ElevadoNodeTree,
};

mod imp {
    use super::*;

    #[derive(Debug, Default, gtk::CompositeTemplate)]
    #[template(resource = "/com/feaneron/Elevado/window.ui")]
    pub struct ElevadoWindow {
        #[template_child]
        pub(super) sort_model: TemplateChild<gtk::SortListModel>,
        #[template_child]
        pub(super) split_view: TemplateChild<adw::OverlaySplitView>,
        #[template_child]
        pub(super) details_sidebar: TemplateChild<ElevadoNodeDetails>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for ElevadoWindow {
        type ParentType = adw::ApplicationWindow;
        type Type = super::ElevadoWindow;

        const NAME: &'static str = "ElevadoWindow";

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
            klass.bind_template_callbacks();

            ElevadoNode::ensure_type();
            ElevadoNodeTree::ensure_type();
            ElevadoSidebarRow::ensure_type();
            ElevadoNodeDetails::ensure_type();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for ElevadoWindow {
        fn constructed(&self) {
            self.parent_constructed();
            glib::spawn_future_local(clone!(
                #[weak(rename_to = win)]
                self,
                async move {
                    win.populate_list().await;
                }
            ));
        }
    }
    impl WidgetImpl for ElevadoWindow {}
    impl WindowImpl for ElevadoWindow {}
    impl ApplicationWindowImpl for ElevadoWindow {}
    impl AdwApplicationWindowImpl for ElevadoWindow {}

    #[gtk::template_callbacks]
    impl ElevadoWindow {
        async fn populate_list(&self) {
            let reg_acc_prox = AccessibleProxy::builder(bus().connection())
                .destination("org.a11y.atspi.Registry")
                .expect("Registry service should exist")
                .path("/org/a11y/atspi/accessible/root")
                .expect("accessible/root object should exist in registry")
                .build()
                .await
                .expect("all objects should implement accessible");

            let objs = reg_acc_prox.get_children().await.unwrap_or_default();
            let mut roots: Vec<ElevadoNode> = vec![];
            for each in objs {
                let node = ElevadoNode::new(each).await;
                // Filter out ourselves to prevent reentrancy
                if let Some(name) = node.name() {
                    if !name.contains("elevado") {
                        roots.push(node);
                    }
                }
            }

            let store = gio::ListStore::new::<ElevadoNode>();
            store.extend_from_slice(&roots);
            self.sort_model.set_model(Some(&store));
        }

        #[expect(clippy::unused_self)]
        #[template_callback]
        fn navigation_page_title_from_node_cb(&self, name: Option<String>) -> String {
            name.unwrap_or_default()
        }

        #[template_callback]
        fn on_sidebar_item_selected(&self, _: glib::ParamSpec, selection: &gtk::SingleSelection) {
            let item = selection.selected_item();
            if let Some(obj) = item {
                let node = obj
                    .downcast_ref::<ElevadoNode>()
                    .expect("this should be a node");

                self.details_sidebar.set_node(node);
            }
        }

        #[template_callback]
        fn on_node_tree_item_selected(&self, value: Option<ElevadoNode>) {
            if let Some(node) = value {
                self.details_sidebar.set_node(node);
                self.split_view.set_show_sidebar(true);
            } else {
                self.split_view.set_show_sidebar(false);
            }
        }

        #[expect(clippy::unused_self)]
        #[template_callback]
        fn window_title_from_node_cb(&self, value: Option<ElevadoNode>) -> String {
            if let Some(node) = value {
                if let Some(name) = node.name() {
                    // Translators: Don't translate content between braces {}
                    return gettext_f("Elevado — {name}", &[("name", &name)]);
                }
            }
            "Elevado".into()
        }
    }
}

glib::wrapper! {
    pub struct ElevadoWindow(ObjectSubclass<imp::ElevadoWindow>)
        @extends gtk::Widget, gtk::Window, gtk::ApplicationWindow, adw::ApplicationWindow,
        @implements gio::ActionGroup, gio::ActionMap;
}

impl ElevadoWindow {
    pub fn new<P: IsA<gtk::Application>>(application: &P) -> Self {
        let win: Self = glib::Object::builder()
            .property("application", application)
            .build();

        win
    }
}
